
#ifndef PLS_INTERNAL_PROFILING_PROFILER_H_
#define PLS_INTERNAL_PROFILING_PROFILER_H_

#ifndef PLS_PROFILING_ENABLED
#define PLS_PROFILING_ENABLED false
#endif

#include <memory>
#include <chrono>
#include <vector>

#include <iostream>

#include "dag_node.h"
#include "thread_stats.h"

namespace pls::internal::profiling {
class profiler {
  using clock = std::chrono::steady_clock;
  using measurement_resolution = std::chrono::nanoseconds;
  using display_resolution = std::chrono::microseconds;

  struct profiler_run {
    profiler_run(unsigned num_threads) : start_time_{},
                                         end_time_{},
                                         root_node_{std::make_unique<dag_node>(0)},
                                         per_thread_stats_(num_threads) {}

    clock::time_point start_time_;
    clock::time_point end_time_;
    std::unique_ptr<dag_node> root_node_;
    std::vector<thread_stats> per_thread_stats_;

    static unsigned long m_to_d(unsigned long duration) {
      measurement_resolution measurement_duration{duration};
      return std::chrono::duration_cast<display_resolution>(measurement_duration).count();
    }

    void print_stats(unsigned num_threads) const;
  };

 public:
  profiler(unsigned num_threads) : num_threads_{num_threads},
                                   profiler_runs_() {}

  dag_node *start_profiler_run();
  void stop_profiler_run();

  void stealing_start(unsigned thread_id);
  void stealing_end(unsigned thread_id, bool success);
  void stealing_cas_op(unsigned thread_id);

  dag_node *task_spawn_child(unsigned thread_id, dag_node *parent);
  dag_node *task_sync(unsigned thread_id, dag_node *synced);
  void task_start_running(unsigned thread_id, dag_node *in_node);
  void task_stop_running(unsigned thread_id, dag_node *in_node);
  void task_prepare_stack_measure(unsigned thread_id, char *stack_memory, size_t stack_size);
  void task_finish_stack_measure(unsigned thread_id, char *stack_memory, size_t stack_size, dag_node *in_node);

 private:
  static constexpr char MAGIC_BYTES[] = {'A', 'B', 'A', 'B', 'A', 'B', 'A', 'B'};
  profiler_run &current_run() {
    return profiler_runs_[profiler_runs_.size() - 1];
  }
  thread_stats &thread_stats_for(unsigned thread_id) {
    return current_run().per_thread_stats_[thread_id];
  }

  unsigned num_threads_;
  std::vector<profiler_run> profiler_runs_;
};
}

#endif //PLS_INTERNAL_PROFILING_PROFILER_H_
